/*
	Copyright (C) 2001 Artem V. Andreev (artem@AA5779.spb.edu)

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

*/
#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#ifdef HAVE_STAT
#include <sys/types.h>
#include <sys/stat.h>
#endif
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#ifdef HAVE_GLOB
#ifdef HAVE_GLOB_H
#include <glob.h>
#endif
#endif
#ifdef HAVE_UNISTD_H
#include <fcntl.h>
#include <unistd.h>
#endif
#include <errno.h>
#include <string.h>
#include <limits.h>
#include <modules.h>
#include <util.h>
#include <dstring.h>
#include <plist.h>
#include "version.h"
/// \section[NAME]
// \em[files] -- a module for NSL (see \ref[sl(1)])

/// \section[OVERVIEW]
// \em[files] is a NSL module which provide interface to file system handling functions
//
// Since most of the builtins are directly bound to corr. system functions/calls, 
// they may or may not present in a certain configuration, so one should generally
// check their presence with \em[feature] before using.
// 
// The following builtins, however, are guaranteed to exist:
// rmfile, mvfile, tmpfile, searchpath, ioerror, clearerr.
// However, the way how these builtins work is highly system-dependent.
// 
//
// For most part, the builtins in the module do not throw errors; instead, they put
// error condition descriptions in an internal buffer accessible via \em2[ioerror][.]
// 



static module *this_mod;
extern procstack *get_pstack(void);
extern storage *create_storage(interpreter *i, storage_class *class, char *name, char *arg);

void files_LTX_init(module *this)
{
	this_mod = this;
}

/// \section[BUILTINS]
#ifdef HAVE_ACCESS
/// exists(filename)
// \returns true if \em[filename] exists
static void bi_exists(chunk *current, interpreter *i, value *args)
{
	if(!access(DS_BODY(args->val), 0))
		ds_appendstr(get_pstack()->state->buffer, "true");
}
#endif

static dstring *error_desc;

static void set_error(void)
{
	if(error_desc)
		ds_unfix(error_desc);
	error_desc = ds_create(strerror(errno));
}

#ifdef HAVE_CHDIR
/// cd(newdir)
// changes the working directory to \em[newdir]
static void bi_cd(chunk *current, interpreter *i, value *args)
{
	if(chdir(DS_BODY(args->val)))
		set_error();
}
#endif

#ifdef HAVE_GETCWD
// getcwd()
// \returns current working directory
// \error[path too long for getcwd]
static void bi_getcwd(chunk *current, interpreter *i, value *args)
{
	static char buf[PATH_MAX + 1];
	if(!getcwd(buf, PATH_MAX))
		recover_error("path too long for getcwd");
	ds_appendstr(get_pstack()->state->buffer, buf);
}
#endif
#ifdef HAVE_MKDIR
/// mkdir(pathname)
// creates a directory \em[pathname]
static void bi_mkdir(chunk *current, interpreter *i, value *args)
{
	if(mkdir(DS_BODY(args->val), S_IREAD | S_IWRITE | S_IEXEC))
		set_error();
}
#endif

#ifdef HAVE_RMDIR
/// rmdir(pathname)
// removes a directory \em[pathname]
// \note On most systems \em[pathname] must be empty
static void bi_rmdir(chunk *current, interpreter *i, value *args)
{
	if(rmdir(DS_BODY(args->val)))
		set_error();
}
#endif

/// rmfile(pathname) or unlink(pathname)
// removes a file \em[pathname]
// \note One generally cannot remove a directory with this function,
// use \em[rmdir] instead
static void bi_rmfile(chunk *current, interpreter *i, value *args)
{
	if(remove(DS_BODY(args->val)))
		set_error();
}

/// mvfile(oldname newname)
// renames or moves \em[oldname] to \em[newname]
static void bi_mvfile(chunk *current, interpreter *i, value *args)
{
	if(rename(DS_BODY(args->val), DS_BODY(next_value(args)->val)))
		set_error();
}

#ifdef HAVE_SYMLINK
/// symlink(source dest)
// makes a symbolic link \em[dest] to \em[source]
static void bi_symlink(chunk *current, interpreter *i, value *args)
{
	if(symlink(DS_BODY(args->val), DS_BODY(next_value(args)->val)))
		set_error();
}
#endif
#ifdef HAVE_LINK
/// link(source dest)
// makes a hard link \em[dest] to \em[source]
static void bi_link(chunk *current, interpreter *i, value *args)
{
	if(link(DS_BODY(args->val), DS_BODY(next_value(args)->val)))
		set_error();
}
#endif

/// tmpfile([prefix])
// \returns a unique name suitable for a temporary file with \em[prefix]
// \note no slashes is added to \em[prefix]
// \error[tmpfile failed: description]
#ifdef HAVE_UNISTD_H
static void bi_tmpfile(chunk *current, interpreter *i, value *args)
{
	static int seqno = 0;
	int f = 0;
	dstring *prefix = args ? args->val : ds_create("/tmp/nsl-"), *tmp;
	do
	{
		f = open(DS_BODY(tmp = ds_concat(prefix, ds_fromint(++seqno, 36, 0))), O_CREAT | O_EXCL, 
				 S_IREAD | S_IWRITE);
		if(f < 0 && errno != EEXIST)
			recover_error("tmpfile failed: %s", strerror(errno));
	}
	while(f < 0);
	close(f);
	remove(DS_BODY(tmp));
	ds_append(get_pstack()->state->buffer, tmp);
}
#endif

#ifdef HAVE_STAT
/// filesize(pathname)
// \returns the size in bytes of \em[pathname]
static void bi_filesize(chunk *current, interpreter *i, value *args)
{
	struct stat st;
	if(stat(DS_BODY(args->val), &st))
		set_error();
	else
		ds_append(get_pstack()->state->buffer, ds_fromint(st.st_size, 10, 0));
}

/// isnewer(pathname1 pathname2)
// \returns true if \em[pathname1] is newer than \em[pathname2]
static void bi_isnewer(chunk *current, interpreter *i, value *args)
{
	struct stat st1, st2;
	if(stat(DS_BODY(args->val), &st1) || stat(DS_BODY(next_value(args)->val), &st2))
		recover_error("can't get file info");
	if(st1.st_mtime > st2.st_mtime)
		ds_appendstr(get_pstack()->state->buffer, "true");
}
#endif

#ifdef HAVE_GLOB
/// glob(pattern [options])
// returns a list containing filenames matching \em[pattern]
// \em[options] may be a combination of the following:
// \item \em[nosort] do not sort the list
// \item \em[noescape] ignore \\ as an escape character in \em[pattern]
// \item \em[nocheck] return \em[pattern] if no matches found
// \item \em[mark] mark directories with trailing /
// \item \em[failonerror] stops at a read error
// \endlist
// \note This builtin's behaviour (inc. the valid patterns) 
// completely depends on the system glob implementation
static void bi_glob(chunk *current, interpreter *i, value *args)
{
	glob_t globber;
	int flags = 0, result;
	char **iter;
	storage *s;
	dstring *null = ds_create(NULL);
	dstring *buffer = ds_create(NULL);
	if(next_value(args))
	{
		char *opts = DS_BODY(next_value(args)->val);
		if(strstr(opts, "nosort"))
			flags |= GLOB_NOSORT;
		if(strstr(opts, "noescape"))
#ifdef GLOB_NOESCAPE		
			flags |= GLOB_NOESCAPE;
#else
			recover_error("your libc doesn't define GLOB_NOESCAPE");
#endif			
		if(strstr(opts, "nocheck"))
			flags |= GLOB_NOCHECK;
		if(strstr(opts, "mark"))
			flags |= GLOB_MARK;
		if(strstr(opts, "failonerror"))
			flags |= GLOB_ERR;
	}
	result = glob(DS_BODY(args->val), flags, NULL, &globber);
#ifdef GLOB_ABORTED	
	if(result == GLOB_ABORTED || result == GLOB_NOSPACE)
#else		
	if(result == GLOB_ABEND || result == GLOB_NOSPACE)
#endif		
	{
		set_error();
		return;
	}
	s = create_storage(i, &plist_storage_class, NULL, NULL);
#ifdef GLOB_NOMATCH	
	if(result != GLOB_NOMATCH)
#else
	if(globber.gl_matchc)
#endif		
	{
		for(iter = globber.gl_pathv; *iter; iter++)
		{
			ds_assign(buffer, NULL);
			ds_appendstr(buffer, *iter);
			s->class->add(i, s, buffer, null);
		}
	}
	globfree(&globber);
}
#endif

/// ioerror()
// \returns a I/O error description produced by other functions of the module
static void bi_ioerror(chunk *current, interpreter *i, value *args)
{
	if(error_desc)
		ds_append(get_pstack()->state->buffer, error_desc);
}

/// clearerr()
// clears the error description to be returned by \em[ioerror]
static void bi_clearerr(chunk *current, interpreter *i, value *args)
{
	if(error_desc)
		ds_unfix(error_desc);
	error_desc = NULL;
}

static void *return_fname(const char *name) { return (void *)name; }

/// searchpath(file pathlist defext [force])
// if \em[file] has no extenstion, append \em[defext] to it
//
// if it does not start with '/', './' or '../', search it in the colon-separated
// \em[pathlist]
//
// if file is not found and \em[force] is present and not empty, places \em[file] 
// at the last path in \em[pathlist]
// \returns constructed filename
static void bi_searchpath(chunk *current, interpreter *i, value *args)
{
	char *name = DS_BODY(args->val);
	char *path = DS_BODY(next_value(args)->val);
	char *defext = DS_BODY(third_value(args)->val);
	ff_ioproc_t crproc = next_value(third_value(args)) && !ds_isempty(next_value(third_value(args))->val) ? return_fname : NULL;
	name = find_filename(name, *path ? path : NULL, *defext ? defext : NULL, NULL, crproc);
	if(name)
		ds_appendstr(get_pstack()->state->buffer, name);
}

#ifdef HAVE_STAT
static int single_mode(char *mode, char **end)
{
	int mask;
	int result = 0;
	switch(*mode)
	{
		case 'u':
			mode++;
			mask = S_IRWXU;
			break;
		case 'g':
			mode++;
			mask = S_IRWXG;
			break;
		case 'o':
			mode++;
			mask = S_IRWXO;
			break;
		default:
			mask = S_IRWXO | S_IRWXU | S_IRWXG;
	
	}
	for(;;)
	{
		switch(*mode)
		{
			case 'w':
				result |= (S_IWUSR | S_IWGRP | S_IWOTH) & mask;
				break;
			case 'r':
				result |= (S_IRUSR | S_IRGRP | S_IROTH) & mask;
				break;
			case 'x':
				result |= (S_IXUSR | S_IXGRP | S_IXOTH) & mask;
				break;
			case 's':
				if(mask == S_IRWXG)
					result |= S_ISGID;
				else if(mask == S_IRWXU)
					result |= S_ISUID;
				break;
			case 't':
				result |= S_ISVTX;
				break;
			default:
				*end = mode;
				return result;
		}
		mode++;
	}
}

static int symbolic_to_numeric(char *mode)
{
	int m = 0;
	char *end;
	while(*mode)
	{
		m |= single_mode(mode, &end);
		if(end == mode)
			break;
		mode = end;
	}
	return m;
}

#ifdef HAVE_UMASK
static void bi_umask(chunk *current, interpreter *i, value *args)
{
	if(args)
	{
		int mask = isdigit(*DS_BODY(args->val)) ? strtoul(DS_BODY(args->val), NULL, 0) : 
					symbolic_to_numeric(DS_BODY(args->val));
		mask = umask(mask);
		ds_append(get_pstack()->state->buffer, ds_fromint(mask, 8, 0));
	}
	else
	{
		int mask = umask(0);
		umask(mask);
		ds_append(get_pstack()->state->buffer, ds_fromint(mask, 8, 0));
	}
}
#endif
	
#endif

static master_instruction_src files_src_add[] = {
#ifdef HAVE_ACCESS	
										{"exists", 1, bi_exists, 1, 1, 1},
#endif
#ifdef HAVE_CHDIR
										{"cd", 0, bi_cd, 1, 1, 1},
#endif
#ifdef HAVE_GETCWD
										{"getcwd", 1, bi_getcwd, 0, 0, 0},
#endif
#ifdef HAVE_MKDIR
										{"mkdir", 0, bi_mkdir, 1, 1, 1},
#endif
										{"rmfile", 0, bi_rmfile, 1, 1, 1},
										{"unlink ", 0, bi_rmfile, 1, 1, 1},
										{"mvfile", 0, bi_mvfile, 2, 2, 2},
#ifdef HAVE_SYMLINK										
										{"symlink", 0, bi_symlink, 2, 2, 2},
#endif									
#ifdef HAVE_LINK
										{"link", 0, bi_link, 2, 2, 2},
#endif
#ifdef HAVE_RMDIR										
										{"rmdir", 0, bi_rmdir, 1, 1, 1},
#endif									
#ifdef HAVE_UNISTD_H
										{"tmpfile", 1, bi_tmpfile, 0, 1, 0},
#endif										
#ifdef HAVE_STAT
										{"filesize", 1, bi_filesize, 1, 1, 1},
										{"isnewer", 1, bi_isnewer, 2, 2, 2},
#endif									
#ifdef HAVE_GLOB
										{"glob", 1, bi_glob, 1, 2, 1},
#endif										
										{"ioerror", 1, bi_ioerror, 0, 0, 0}, 
										{"clearerr", 0, bi_clearerr, 0, 0, 0}, 
										{"searchpath", 1, bi_searchpath, 3, 4, 3},
#ifdef HAVE_UMASK
										{"umask", 1, bi_umask, 0, 1, 1},
#endif				
										{NULL}
										};

void files_LTX_attach(mod_attach *ma)
{
//	ma->data = xmalloc(sizeof(files_src) * sizeof(instruction_t) / sizeof(*files_src));
//	replace_builtins(ma->i, files_src, ma->data);
	append_builtins(ma->i, files_src_add);
}

const char files_LTX_version[] = VERSION;
const char files_LTX_date[] = __DATE__;
/// \section[AUTHOR]
// Artem V. Andreev

/// \section[COPYRIGHT]
// see \ref[sl(1)]

/// \seealso
