#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include <pgsql/libpq-fe.h>
#include <modules.h>
#include <util.h>
#include <lists.h>
#include <dstring.h>
#include "version.h"

static module *this_mod;
extern procstack *get_pstack(void);

static void cleanup_pg(void);
void postgres_LTX_init(module *this)
{
	this_mod = this;
	enh_atexit(cleanup_pg);
}

DEFLIST(pgsql_engine_data,
	unsigned char is_active : 1;
	instruction_t chain;
	PGconn *connection;
	char *error_msg;
	dstring *response;
	interpreter *parent;
);

static pgsql_engine_data *all_engines;

typedef char *(*sql_api_t)(pgsql_engine_data *inst, char *arg);

static void sql_response_handler(void *data, const char *message)
{
	pgsql_engine_data *psql = data;
	if(!psql->response)
		psql->response = ds_fix(ds_create(NULL));
	ds_appendstr(psql->response, message);
}

static void cleanup_pg(void)
{
	pgsql_engine_data *eng = all_engines;
	for(; eng; eng = list_next(eng))
	{
		if(eng->connection)
			PQfinish(eng->connection);
	}
}

static char *sql_connect(pgsql_engine_data *inst, char *arg)
{
	if(inst->connection)
	{
		inst->error_msg = "SL/SQL: connection already established";
		return "";
	}
	if(!arg)
		recover_error("SL/SQL (connect): missing argument");
	inst->connection = PQconnectdb(arg);
	if(PQstatus(inst->connection) != CONNECTION_OK)
	{
		inst->error_msg = "SL/SQL: cannot connect to a database";
		PQfinish(inst->connection);
		inst->connection = NULL;
		return "";
	}
	PQsetNoticeProcessor(inst->connection, sql_response_handler, inst);
	return "true";
}

static char *sql_disconnect(pgsql_engine_data *inst, char *arg)
{
	if(!inst->connection)
	{
		inst->error_msg = "SL/SQL: nothing to disconnect";
		return "";
	}
	PQfinish(inst->connection);
	inst->connection = NULL;
	return "true";
}

static char *sql_engine(pgsql_engine_data *inst, char *arg)
{
	return "PostgreSQL";
}

static table_t sql_entries[];
static char *sql_feature(pgsql_engine_data *inst, char *arg)
{
	if(!arg)
	{
		inst->error_msg = "SL/SQL (feature): missing argument";
		return NULL;
	}
	return lookup_table(sql_entries, arg) ? "true" : "";
}

static char *sql_error(pgsql_engine_data *inst, char *arg)
{
	return inst->error_msg ? inst->error_msg : "";
}

static char *sql_response(pgsql_engine_data *inst, char *arg)
{
	return inst->response ? DS_BODY(inst->response) : "";
}

static char *sql_clearerr(pgsql_engine_data *inst, char *arg)
{
	inst->error_msg = NULL;
	if(inst->response)
		ds_unfix(inst->response);
	inst->response = NULL;
	return "";
}

static char *sql_exec(pgsql_engine_data *inst, char *arg)
{
	PGresult *result;
	dstring *count;
	
	if(!inst->connection)
	{
		inst->error_msg = "SL/SQL: no connection established";
		return "";
	}
	result = PQexec(inst->connection, arg);
	
	inst->error_msg = PQerrorMessage(inst->connection);
	if(inst->error_msg && *inst->error_msg)
		count = ds_create(NULL);
	else
	{
		if(PQntuples(result))
			count = ds_fromint(PQntuples(result), 10, 0);
		else if(*PQcmdTuples(result))
			count = ds_create(PQcmdTuples(result));
		else
			count = ds_create("0");
	}
	if(result)
		PQclear(result);
	return DS_BODY(count);
}

static char *sql_unique_name(void)
{
	static int counter = 1;
	static char buffer[64];
	sprintf(buffer, "#sql:%8.8X", counter++);
	return xstrdup(buffer);
}

static storage *sql_create_dummy(interpreter *i, char *name, char *arg)
{
	fatal_error("SL/SQL: internal error");
	return NULL; // does nothing just inserted to shut up the compiler
}

static void sql_update_dummy(interpreter *i, storage *s, dstring *key, dstring *val)
{
	// dummy
}

static void sql_remove_dummy(interpreter *i, storage *s, dstring *key)
{
	// dummy
}

static void sql_enditer_dummy(interpreter *i, storage *s, void **data)
{
	// dummy
}

static void sql_query_destroy(interpreter *i, storage *s)
{
	PQclear(s->data);
}

typedef struct query_row_data
{
	PGresult *res;
	int tno;
} query_row_data;

static void sql_row_destroy(interpreter *i, storage *s)
{
	FREE((query_row_data *)s->data);
}

static dstring *sql_row_get(interpreter *i, storage *s, dstring *key)
{
	query_row_data *qrd = s->data;
	int fno = PQfnumber(qrd->res, DS_BODY(key));
	char *data;
	if(fno < 0)
		recover_error("SL/SQL: unknown field name '%s'", DS_BODY(key));
	data = PQgetvalue(qrd->res, qrd->tno, fno);
	return ds_create(data);
}

static int sql_row_check(interpreter *i, storage *s, dstring *key)
{
	query_row_data *qrd = s->data;
	int ky = convert_to_int(DS_BODY(key));
	return ky >= 0 && ky < PQnfields(qrd->res);
}

static dstring *sql_row_firstitem(interpreter *i, storage *s, void **data)
{
	query_row_data *qrd = s->data;
	if(!PQnfields(qrd->res))
		return NULL;
	*data = (void *)1;
	return ds_create(PQfname(qrd->res, 0));
}

static dstring *sql_row_nextitem(interpreter *in, storage *s, void **data)
{
	query_row_data *qrd = s->data;
	dstring *ds;
	int i = (int)*data;
	if(i == PQnfields(qrd->res))
		return NULL;
	ds = ds_create(PQfname(qrd->res, i++));
	*data = (void *)i;
	return ds;
}

static int sql_row_numofitems(interpreter *i, storage *s)
{
	return PQnfields(((query_row_data *)s->data)->res);
}

static storage_class sql_row_class = {sql_create_dummy, sql_row_destroy,
									  sql_row_check, sql_row_get, 
									  sql_update_dummy, sql_update_dummy,
									  sql_remove_dummy, sql_row_firstitem,
									  sql_row_nextitem, sql_enditer_dummy,
									  sql_row_numofitems
									 };

static dstring *sql_query_get(interpreter *i, storage *s, dstring *key)
{
	int ky = convert_to_int(DS_BODY(key));
	storage *sto = ALLOC(storage);
	query_row_data *qrd = ALLOC(query_row_data);
	char *name;
	
	if(ky < 0 || ky >= PQntuples(s->data))
		recover_error("SL/SQL: invalid index %d", ky);
	qrd->res = s->data;
	qrd->tno = ky;
	sto->data = qrd;
	sto->class = &sql_row_class;
	add_shared_int(i, init_shared_storage(i, sto, name = sql_unique_name()));
	return ds_create(name);
}

static int sql_query_check(interpreter *i, storage *s, dstring *key)
{
	int ky = convert_to_int(DS_BODY(key));
	return ky >= 0 && ky < PQntuples(s->data);
}

static dstring *sql_query_firstitem(interpreter *i, storage *s, void **data)
{
	if(!PQntuples(s->data))
		return NULL;
	else
	{
		*data = (void *)1;
		return ds_create("0");
	}
}

static dstring *sql_query_nextitem(interpreter *in, storage *s, void **data)
{
	int i = (int)*data;
	dstring *ds;
	if(i == PQntuples(s->data))
		return NULL;
	ds = ds_fromint(i++, 10, 0);
	*data = (void *)i;
	return ds;
}

static int sql_query_numofitems(interpreter *i, storage *s)
{
	return PQntuples(s->data);
}

static storage_class sql_query_class = {sql_create_dummy, sql_query_destroy,
										sql_query_check, sql_query_get,
										sql_update_dummy, sql_update_dummy,
										sql_remove_dummy, sql_query_firstitem,
										sql_query_nextitem, sql_enditer_dummy,
										sql_query_numofitems
										};

static char *sql_query(pgsql_engine_data *inst, char *query)
{
	PGresult *result;
	storage *sto;
	char *name;
	if(!inst->connection)
	{
		inst->error_msg = "SL/SQL: no connection established";
		return "";
	}
	result = PQexec(inst->connection, query);
	inst->error_msg = PQerrorMessage(inst->connection);
	if(inst->error_msg && *inst->error_msg)
		return "";
	sto = ALLOC(storage);
	sto->class = &sql_query_class;
	sto->data = result;
	add_shared_int(inst->parent, init_shared_storage(inst->parent, sto, name = sql_unique_name()));
	return name;
}

static table_t sql_entries[] = {{"connect", sql_connect}, 
								{"disconnect", sql_disconnect},
								{"query", sql_query},
								{"exec", sql_exec},
								{"engine", sql_engine},
								{"error", sql_error},
								{"response", sql_response},
								{"feature", sql_feature},
								{"clearerr", sql_clearerr},
								{NULL, NULL}
							   };

void sql_interface(chunk *current, interpreter *i, value *args)
{
	pgsql_engine_data *data = get_ma_data(this_mod, i);
	sql_api_t api;
	if(!data->is_active)
	{
		if(!ds_comparestr(args->val, "engine", NULL, NULL) && 
			next_value(args) && 
			!ds_comparestr(next_value(args)->val, "postgresql", ds_p_casefold, NULL)
		  )
		{
			data->is_active = 1;
			ds_appendstr(get_pstack()->state->buffer, "PostgreSQL");
		}
		else
			data->chain(current, i, args);
		return;
	}
	api = lookup_table(sql_entries, DS_BODY(args->val));
	if(!api)
		data->error_msg = "SL/SQL: undefined request";
	else 
	{
		char *res = api(data, list_next(args) ? DS_BODY(next_value(args)->val) : NULL);
		if(res)
			ds_appendstr(get_pstack()->state->buffer, res);
	}
}


static master_instruction_src postgres_src[] = {
										 {"sql", 1, sql_interface, 1, 2, 2},
										 {NULL}
										};

void postgres_LTX_attach(mod_attach *ma)
{
	ma->data = ALLOC(pgsql_engine_data);
	((pgsql_engine_data *)ma->data)->is_active = 1;
	((pgsql_engine_data *)ma->data)->parent = ma->i;
	all_engines = list_add(all_engines, ma->data);
	replace_builtins(ma->i, postgres_src, &((pgsql_engine_data *)ma->data)->chain);
}

const char postgres_LTX_version[] = VERSION;
const char postgres_LTX_date[] = __DATE__;

